@@ -7,7 +7,7 @@ import logging
 import functools
 from psycopg2 import sql
 from psycopg2 import ProgrammingError, IntegrityError
-from psycopg2.errorcodes import UNIQUE_VIOLATION
+from psycopg2.errorcodes import UNDEFINED_COLUMN, UNIQUE_VIOLATION
 from psycopg2.extensions import AsIs
 from .openupgrade import logged_query, version_info
 from .openupgrade_tools import column_exists, table_exists
@@ -17,7 +17,7 @@ logger.setLevel(logging.DEBUG)
 
 
 def _change_foreign_key_refs(env, model_name, record_ids, target_record_id,
-                             exclude_columns, model_table):
+                             exclude_columns, model_table, extra_where=None):
     # As found on https://stackoverflow.com/questions/1152260
     # /postgres-sql-to-list-table-foreign-keys
     # Adapted for specific Odoo structures like many2many tables
@@ -39,20 +39,27 @@ def _change_foreign_key_refs(env, model_name, record_ids, target_record_id,
         # Try one big swoop first
         env.cr.execute('SAVEPOINT sp1')  # can't use env.cr.savepoint() in base
         try:
+            query = sql.SQL(
+                """UPDATE {table}
+                SET {column} = %(target_record_id)s
+                WHERE {column} in %(record_ids)s"""
+            ).format(
+                table=sql.Identifier(table), column=sql.Identifier(column),
+            )
+            if extra_where:
+                query += sql.SQL(extra_where)
             logged_query(
-                env.cr, """UPDATE %(table)s
-                SET "%(column)s" = %(target_record_id)s
-                WHERE "%(column)s" in %(record_ids)s
-                """, {
-                    'table': AsIs(table),
-                    'column': AsIs(column),
+                env.cr, query, {
                     'record_ids': record_ids,
                     'target_record_id': target_record_id,
                 }, skip_no_result=True,
             )
         except (ProgrammingError, IntegrityError) as error:
             env.cr.execute('ROLLBACK TO SAVEPOINT sp1')
-            if error.pgcode != UNIQUE_VIOLATION:
+            if error.pgcode == UNDEFINED_COLUMN and extra_where:
+                # extra_where is introducing a bad column. Ignore this table.
+                continue
+            elif error.pgcode != UNIQUE_VIOLATION:
                 raise
             # Fallback on setting each row separately
             m2m_table = not column_exists(env.cr, table, 'id')
@@ -293,21 +300,115 @@ def _change_translations_sql(env, model_name, record_ids, target_record_id,
         })
 
 
+def apply_operations_by_field_type(
+        env, model_name, record_ids, target_record_id, field_spec, field_vals,
+        field_type, column, operation, method):
+    vals = {}
+    o2m_changes = 0
+    if method == 'orm':
+        model = env[model_name]
+        all_records = model.browse((target_record_id,) + tuple(record_ids))
+        target_record = model.browse(target_record_id)
+        first_value = getattr(target_record, column)
+        field = model._fields[column]
+    else:
+        first_value = field_vals[0]
+    if field_type in ('char', 'text', 'html'):
+        if not operation:
+            operation = 'other' if field_type == 'char' else 'merge'
+        if operation == 'first_not_null':
+            field_vals = [x for x in field_vals if x]
+            if field_vals:
+                vals[column] = field_vals[0]
+        elif operation == 'merge':
+            _list = filter(lambda x: x, field_vals)
+            vals[column] = ' | '.join(_list)
+    elif field_type in ('integer', 'float', 'monetary'):
+        if operation or field_type != 'integer':
+            field_vals = [0 if not x else x for x in field_vals]
+        if not operation:
+            operation = 'other' if field_type == 'integer' else 'sum'
+        if operation == 'sum':
+            vals[column] = sum(field_vals)
+        elif operation == 'avg':
+            vals[column] = sum(field_vals) / len(field_vals)
+        elif operation == 'max':
+            vals[column] = max(field_vals)
+        elif operation == 'min':
+            vals[column] = min(field_vals)
+    elif field_type == 'boolean':
+        if operation:
+            field_vals = [False if x is None else x for x in field_vals]
+        operation = operation or 'other'
+        if operation == 'and':
+            vals[column] = functools.reduce(lambda x, y: x & y, field_vals)
+        elif operation == 'or':
+            vals[column] = functools.reduce(lambda x, y: x | y, field_vals)
+    elif field_type in ('date', 'datetime'):
+        if operation:
+            field_vals = list(filter(lambda x: x, field_vals))
+        operation = field_vals and operation or 'other'
+        if operation == 'max':
+            vals[column] = max(field_vals)
+        elif operation == 'min':
+            vals[column] = min(field_vals)
+    elif field_type == 'many2many' and method == 'orm':
+        operation = operation or 'merge'
+        if operation == 'merge':
+            field_vals = filter(lambda x: x is not False, field_vals)
+            vals[column] = [(4, x.id) for x in field_vals]
+    elif field_type == 'one2many' and method == 'orm':
+        operation = operation or 'merge'
+        if operation == 'merge':
+            o2m_changes += 1
+            field_vals.write({field.inverse_name: target_record_id})
+    elif field_type == 'binary':
+        operation = operation or 'merge'
+        if operation == 'merge':
+            field_vals = [x for x in field_vals if x]
+            if not first_value and field_vals:
+                vals[column] = field_vals[0]
+    elif field_type in ('many2one', 'reference'):
+        operation = operation or 'merge'
+        if operation == 'merge':
+            if method != 'orm':
+                field_vals = [x for x in field_vals if x]
+            if not first_value and field_vals:
+                vals[column] = field_vals[0]
+    elif field_type == 'many2one_reference' and method == 'orm' and\
+            field.model_field in model._fields:
+        operation = operation or 'merge'
+        if operation == 'merge':
+            if field.model_field in field_spec:
+                del field_spec[field.model_field]
+            list_model_field = all_records.mapped(field.model_field)
+            zip_list = [(x, y) for x, y
+                        in zip(field_vals, list_model_field) if x and y]
+            if first_value and zip_list:
+                vals[column] = zip_list[0][0]
+                vals[field.model_field] = zip_list[0][1]
+    if method == 'orm':
+        return vals, o2m_changes
+    else:
+        return vals
+
+
 def _adjust_merged_values_orm(env, model_name, record_ids, target_record_id,
                               field_spec):
     """This method deals with the values on the records to be merged +
-    the target record, performing operations that makes sense on the meaning
+    the target record, performing operations that make sense on the meaning
     of the model.
 
-    :param field_spec: Dictionary with field names as keys and forced operation
-      to perform as values. If a field is not present here, default operation
-      will be performed.
+    :param: field_spec: Dictionary with field names as keys and forced
+      operation to perform as values. If a field is not present here, default
+      operation will be performed.
 
       Possible operations by field types:
 
       * Char, Text and Html fields:
         - 'merge' (default for Text and Html): content is concatenated
           with an ' | ' as separator
+        - 'first_not_null': Put first not null value.
         - other value (default for Char): content on target record is preserved
       * Integer, Float and Monetary fields:
         - 'sum' (default for Float and Monetary): Sum all the values of
@@ -339,14 +440,22 @@ def _adjust_merged_values_orm(env, model_name, record_ids, target_record_id,
       * One2many fields:
         - 'merge' (default): combine all the values
         - other value: content on target record is preserved
+      * Many2manyReference fields:
+        - 'merge' (default): if its model_field is in field_spec,
+        delete it from there. Apply first positive (on field and
+        corresponding model_field) of the records if value of target record
+        is not positive, preserve target value otherwise.
+        - other value: content on target record is preserved
       * Reference fields:
-        - any value: content on target record is preserved
+        - 'merge' (default): apply first not null value of the records if
+        value of target record is null, preserve target value otherwise.
+        - other value: content on target record is preserved
       * Selection fields:
         - any value: content on target record is preserved
     """
     model = env[model_name]
     fields = model._fields.values()
-    all_records = model.browse(tuple(record_ids) + (target_record_id, ))
+    all_records = model.browse((target_record_id,) + tuple(record_ids))
     target_record = model.browse(target_record_id)
     vals = {}
     o2m_changes = 0
@@ -354,59 +463,17 @@ def _adjust_merged_values_orm(env, model_name, record_ids, target_record_id,
         if not field.store or field.compute or field.related:
             continue  # don't do anything on these cases
         op = field_spec.get(field.name, False)
-        _list = all_records.mapped(field.name)
-        if field.type in ('char', 'text', 'html'):
-            if not op:
-                op = 'other' if field.type == 'char' else 'merge'
-            if op == 'merge':
-                _list = filter(lambda x: x is not False, _list)
-                vals[field.name] = ' | '.join(_list)
-        elif field.type in ('integer', 'float', 'monetary'):
-            if not op:
-                op = 'other' if field.type == 'integer' else 'sum'
-            if op == 'sum':
-                vals[field.name] = sum(_list)
-            elif op == 'avg':
-                vals[field.name] = sum(_list) / len(_list)
-            elif op == 'max':
-                vals[field.name] = max(_list)
-            elif op == 'min':
-                vals[field.name] = min(_list)
-        elif field.type == 'boolean':
-            op = op or 'other'
-            if op == 'and':
-                vals[field.name] = functools.reduce(lambda x, y: x & y, _list)
-            elif op == 'or':
-                vals[field.name] = functools.reduce(lambda x, y: x | y, _list)
-        elif field.type in ('date', 'datetime'):
-            if op:
-                _list = filter(lambda x: x is not False, _list)
-            op = _list and op or 'other'
-            if op == 'max':
-                vals[field.name] = max(_list)
-            elif op == 'min':
-                vals[field.name] = min(_list)
-        elif field.type == 'many2many':
-            op = op or 'merge'
-            if op == 'merge':
-                _list = filter(lambda x: x is not False, _list)
-                vals[field.name] = [(4, x.id) for x in _list]
-        elif field.type == 'one2many':
-            op = op or 'merge'
-            if op == 'merge':
-                o2m_changes += 1
-                _list.write({field.inverse_name: target_record_id})
-        elif field.type == 'binary':
-            op = op or 'merge'
-            if op == 'merge':
-                _list = [x for x in _list if x]
-                if not getattr(target_record, field.name) and _list:
-                    vals[field.name] = _list[0]
-        elif field.type == 'many2one':
-            op = op or 'merge'
-            if op == 'merge':
-                if not getattr(target_record, field.name) and _list:
-                    vals[field.name] = _list[0]
+        if field.type != 'reference':
+            _list = all_records.mapped(field.name)
+        else:
+            _list = [x[field.name] for x in all_records if x[field.name]]
+        field_vals, field_o2m_changes = apply_operations_by_field_type(
+            env, model_name, record_ids, target_record_id, field_spec,
+            _list, field.type, field.name, op, 'orm')
+        vals.update(field_vals)
+        o2m_changes += field_o2m_changes
+    if not vals:
+        return
     # Curate values that haven't changed
     new_vals = {}
     for f in vals:
@@ -424,6 +491,82 @@ def _adjust_merged_values_orm(env, model_name, record_ids, target_record_id,
         )
 
 
+def _adjust_merged_values_sql(env, model_name, record_ids, target_record_id,
+                              model_table, field_spec):
+    """This method deals with the values on the records to be merged +
+    the target record, performing operations that make sense on the meaning
+    of the model.
+
+    :param: field_spec: Dictionary with field names as keys and forced
+      operation to perform as values. If a field is not present here, default
+      operation will be performed.
+
+      Possible operations by field types same as _adjust_merged_values_orm.
+    """
+    if not column_exists(env.cr, model_table, 'id'):
+        # TODO: handle one2many and many2many
+        return
+    env.cr.execute("""
+        SELECT isc.column_name, isc.data_type, imf.ttype
+        FROM information_schema.columns isc
+        JOIN ir_model_fields imf ON (
+            imf.name = isc.column_name AND imf.model = %s)
+        WHERE isc.table_name = %s
+        """, (model_name, model_table))
+    dict_column_type = env.cr.fetchall()
+    columns = ', '.join([x[0] for x in dict_column_type])
+    env.cr.execute(
+        """SELECT {columns}
+        FROM {table}
+        WHERE id IN %(record_ids)s""".format(
+            table=model_table,
+            columns=columns,
+        ), {'record_ids': (target_record_id,) + tuple(record_ids)}
+    )
+    lists = list(zip(*(env.cr.fetchall())))
+    new_vals = {}
+    vals = {}
+    for i, (column, column_type, field_type) in enumerate(dict_column_type):
+        op = field_spec.get(column, False)
+        _list = list(lists[i])
+        field_vals = apply_operations_by_field_type(
+            env, model_name, record_ids, target_record_id, field_spec,
+            _list, field_type, column, op, 'sql')
+        vals.update(field_vals)
+    if not vals:
+        return
+    # Curate values that haven't changed
+    env.cr.execute(
+        """SELECT {columns}
+        FROM {table}
+        WHERE id = %(target_record_id)s
+        """.format(
+            table=model_table,
+            columns=", ". join(list(vals.keys()))
+        ), {'target_record_id': target_record_id}
+    )
+    record_vals = env.cr.dictfetchall()
+    for column in vals:
+        if vals[column] != record_vals[0]:
+            new_vals[column] = vals[column]
+    if new_vals:
+        ident_dict = {x: sql.Identifier(x) for x in new_vals.keys()}
+        query = sql.SQL(
+            "UPDATE {table} SET {set_value} WHERE {id} = %(target_record_id)s"
+        ).format(
+            table=sql.Identifier(model_table),
+            id=sql.Identifier("id"),
+            set_value=sql.SQL(
+                ", ".join([
+                    "{{{field}}} = %({field})s".format(field=x)
+                    for x in new_vals.keys()
+                ])
+            ).format(**ident_dict)
+        )
+        new_vals["target_record_id"] = target_record_id
+        logged_query(env.cr, query, new_vals)
+
+
 def _change_generic(env, model_name, record_ids, target_record_id,
                     exclude_columns, method='orm'):
     """ Update known generic style res_id/res_model references """
@@ -484,15 +627,26 @@ def _change_generic(env, model_name, record_ids, target_record_id,
             query = sql.SQL(
                 """UPDATE {table} SET {res_id_column} = %(target_record_id)s
                 WHERE {model_column} = %(model_name)s
-                AND {res_id_column} in %(record_ids)s
-                """).format(**format_args)
-            if model_to_replace == 'mail.followers':
-                query += sql.SQL("""AND partner_id NOT IN (
-                    SELECT partner_id FROM {table}
-                    WHERE {res_id_column} = %(target_record_id)s
-                )""").format(**format_args)
-            logged_query(env.cr, query, query_args, skip_no_result=True)
-            if model_to_replace == 'mail.followers':
+                """
+            ).format(**format_args)
+            if model_to_replace != 'mail.followers':
+                query += sql.SQL(
+                    "AND {res_id_column} in %(record_ids)s"
+                ).format(**format_args)
+                logged_query(env.cr, query, query_args, skip_no_result=True)
+            else:
+                for record_id in record_ids:
+                    query_args["record_id"] = record_id
+                    query2 = query + sql.SQL(
+                        """AND {res_id_column} = %(record_id)s
+                        AND partner_id NOT IN (
+                            SELECT partner_id FROM {table}
+                            WHERE {res_id_column} = %(target_record_id)s
+                        )"""
+                    ).format(**format_args)
+                    logged_query(
+                        env.cr, query2, query_args, skip_no_result=True,
+                    )
                 # Remove remaining records non updated (that are duplicates)
                 logged_query(env.cr, sql.SQL(
                     "DELETE FROM {table} "
@@ -555,7 +709,7 @@ def merge_records(env, model_name, record_ids, target_record_id,
     """
     if exclude_columns is None:
         exclude_columns = []
-    if field_spec is None:
+    if field_spec is None and method == 'orm':
         field_spec = {}
     if isinstance(record_ids, list):
         record_ids = tuple(record_ids)
@@ -594,7 +748,9 @@ def merge_records(env, model_name, record_ids, target_record_id,
         _change_foreign_key_refs(*args3)
         _change_reference_refs_sql(*args)
         _change_translations_sql(*args)
-        # TODO: Adjust values of the merged records through SQL
+        if field_spec is not None:
+            args4 = args0 + (model_table,) + (field_spec,)
+            _adjust_merged_values_sql(*args4)
         if delete:
             _delete_records_sql(
                 env, model_name, record_ids, target_record_id,
